/*
 * Copyright (C) 1998, 1999  Sonny Parlin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA.
 */

#include <sys/stat.h>
#include <sys/wait.h>
#include "np.h"

#define MAXFD      		64
#define TRANSFER_BUFFER_SIZE	10000
#define LISTENQ			1024

/*
 * argv[1] = source port;
 * argv[2] = destination server;
 * argv[3] = desintation port;
 * argv[4] = source IP;
 */

void *Proxy (void *ptr);
int FilterIP(char *);
int proxy (int sd, char *szHost, char *szPort);
void sig_chld (int signo);
int daemon_init (const char *pname, int facility);

typedef struct tag_transfer
  {
    int fdSource;
    int fdDest;
  }
Transfer;

struct proxy_t
  {
    char *ip;
    char *port;
    int fd;
  };

/*
 * int main()
 */
int 
main (int argc, char *argv[])
{
  int listenfd, connfd;
  socklen_t clilen, clientaddrlen;
  socklen_t addrlen;
  struct sockaddr_in cliaddr;
  struct sockaddr *clientaddr;
  struct proxy_t *ptr;
  pthread_t ptConnection;
  int FilterCheck;

  if (argc < 4)
    {
      printf ("Usage: proxy [listen port][dest server][dest port][source IP]\n");
      exit (0);
    }

  ptr = (struct proxy_t *) malloc (sizeof (struct proxy_t));

  ptr->ip = argv[2];
  ptr->port = argv[3];

  /* Run proxy as UNIX daemon */
  daemon_init (argv[0], 6);

   if (argc == 5) {
     listenfd = tcp_listen (argv[4], argv[1], &addrlen); 
   } else { 
     listenfd = tcp_listen (NULL, argv[1], &addrlen); 
   }

  for (;;)
    {
      clilen = sizeof (cliaddr);
      connfd = accept (listenfd, (SA *) &cliaddr, &clilen);
      if (connfd == -1)
	{
	  continue;
	}

      ptr->fd = connfd;

      if ( (clientaddr = malloc(MAXSOCKADDR)) == NULL) {
        syslog(LOG_ERR, "memory error %m");
      } 

      clientaddrlen = MAXSOCKADDR;
      if (getpeername(ptr->fd, clientaddr, &clientaddrlen) < 0) {
        syslog(LOG_ERR, "getpeername error %m");
      }

      /*
       * Basic IP filtering
       */
      FilterCheck = FilterIP(Sock_ntop(clientaddr, clientaddrlen));
      if (FilterCheck > 0) {

        switch (FilterCheck) {
          case 1:
            syslog(LOG_INFO, "found deny rule for %s, dropping connection!", 
	      Sock_ntop(clientaddr, clientaddrlen));
            break;
          case 2:
            syslog(LOG_INFO, "%s matched default policy of DENY, dropping connection!", 
	      Sock_ntop(clientaddr, clientaddrlen));
            break;
          case 3:
            syslog(LOG_ERR, "No rule matching %s and no default policy, "
			    "dropping!",
	      Sock_ntop(clientaddr, clientaddrlen)); 
        }
        
        if (close(ptr->fd) == -1)
          syslog(LOG_ERR, "Close problems %m");
      } else {
        syslog(LOG_INFO, "allowing connection from %s", 
	  Sock_ntop(clientaddr, clientaddrlen));
      }

      if (pthread_create (&ptConnection,
			  NULL,
			  Proxy,
			  (void *) ptr) != 0)
	{
          syslog(LOG_ERR, "Thread creation problems %m");
	}
    }
}

/*
 * Proxy()
 *
 * Extracts function parameters from ptr used by pthread_create
 * and calls proxy().
 */
void *
Proxy (void *ptr)
{
  int connfd;
  struct proxy_t *pt;
  
  pt = (struct proxy_t *) ptr;
  connfd = pt->fd;
  proxy (connfd, pt->ip, pt->port);
  return NULL;
} 

/*
 * void proxy()
 *
 * Reads data from socket descriptor created by new client 
 * and passes it to socket descriptor we create. Synchronous
 * I/O multiplexing is done via select. See SELECT(2) for details.
 */
int proxy (int sd, char *szHost, char *szPort)
{
  int sdOut;
  char szBuffer[TRANSFER_BUFFER_SIZE];
  int Readval, Writeval;
  fd_set rset; 
  int selectval;
  struct sockaddr_in to_addr;
  
  /*
   * Connect to requested host
   */
  sdOut = tcp_connect (szHost, szPort);
  
#ifdef DEBUG
  printf("%s:%s\n\n", hostname, service);
#endif

  /*
   * Set up a select loop for reading and writing
   */

  for (;;)
    {
      FD_SET(sd, &rset);
      FD_SET(sdOut, &rset);
      selectval = select(1024, &rset, NULL, NULL, NULL);
      if (selectval == -1) {
        syslog(LOG_ERR, "select(): %m");
        return 1;
      } 
      
      /*
       * Is client readable? If so, read from client, 
       * write to destination.
       */
      if(FD_ISSET(sd, &rset)) {
        Readval = read (sd, &szBuffer, TRANSFER_BUFFER_SIZE);
        
        if (Readval <= 0)
          {
            if (close (sdOut) == -1)
              syslog(LOG_ERR, "Close problems %m");
            if (close (sd) == -1)
              syslog(LOG_ERR, "Close problems %m");
            return EXIT_SUCCESS;
          } 
        while (Readval > 0)
          {
            Writeval = write (sdOut, &szBuffer, Readval);
            if (Writeval <= 0)
              return EXIT_SUCCESS;
              
            Readval -= Readval;
          } 
      }

      /*
       * Is destination readable? If so, read from destination, 
       * write to client. 
       */
      if(FD_ISSET(sdOut, &rset)) {
        Readval = read (sdOut, &szBuffer, TRANSFER_BUFFER_SIZE);
        
        if (Readval <= 0)
          {
            if (close (sd) == -1)
              syslog(LOG_ERR, "Close problems %m");
            if (close (sdOut) == -1)
              syslog(LOG_ERR, "Close problems %m");
            return EXIT_SUCCESS;
          } 
        while (Readval > 0)
          {
            Writeval = write (sd, &szBuffer, Readval);
            if (Writeval <= 0)
              return EXIT_SUCCESS;
              
            Readval -= Readval;
          } 
      }   
    } 
  return EXIT_SUCCESS;
}

/*
 * int daemon_init()
 *
 * Causes proxy to run as a UNIX daemon
 */
int 
daemon_init (const char *pname, int facility)
{
  int i;
  pid_t pid;

  if ((pid = fork ()) != 0)
    exit (0);

  setsid ();

  signal (SIGHUP, SIG_IGN);

  if ((pid = fork ()) != 0)
    exit (0);

  chdir ("/");

  umask (0);

  for (i = 0; i < MAXFD; i++)
    close (i);

  openlog (pname, LOG_PID, facility);

  return 0;
}

/*
 * int FilterIP()
 *
 * Basic IP filtering
 */
int FilterIP(char *clientip)
{
  FILE *access;
  char str[MAXLINE];
  int deny = 0, accept = 0;
  char *access_str, *deny_str;
  int i;

  access = fopen("/etc/proxy.filters", "r");

  if (access == NULL)
    return 0;

  access_str = (char *)malloc(sizeof(char) * 25);
  sprintf(access_str, "ACCEPT.%s", clientip);

  deny_str = (char *)malloc(sizeof(char) * 25);
  sprintf(deny_str, "DENY.%s", clientip);

  while (fgets(str, MAXLINE, access) != NULL)
  {
    if (str[0] == '#')
      continue;

    for(i=0;str[i] != '\n'; i++);
    str[strlen(str)-1] = '\0';

    for(i=0;str[i] != '.'; i++)
	str[i] = toupper(str[i]);

    if (strstr(access_str, str) != NULL) {
	free(access_str);
	free(deny_str);
        return 0;
    }

    if (strstr(deny_str, str) != NULL) {
	free(access_str);
	free(deny_str);
        return 1;
    }

    if (strcmp(str, "ACCEPT;") == 0)
      accept++; 
    else if (strcmp(str, "DENY;") == 0)
      deny++;
  } 
  free(access_str);
  free(deny_str);
  fclose(access);

  if (deny > 0)
    return 2;
  else if (accept > 0)
    return 0;
  else
    return 3;
}
